/*
 *  Copyright 1999-2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

/***************************************************************************
 * Description: DSAPI plugin for Lotus Domino                              *
 * Author:      Andy Armstrong <andy@tagish.com>                           *
 * Date:        2007-02-06                                                 *
 * Modified by: Vladimir Anufriev <avn@avnsite.com>                        *
 * Version:     $Revision: 1.13.1 $                                        *
 *									   *
 *									   *
 * History:								   *
 * 2007-03-28	avn							   *
 *	map_uri_to_worker() returns "default=..." instead NULL		   *
 ***************************************************************************/

/* Andy: Based on the IIS redirector by Gal Shachor <shachor@il.ibm.com> */

/* Domino stuff */
#include <global.h>
#include <addin.h>
#include <dsapi.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <pthread.h>
#include <regex.h>

/* JK stuff */
#include "jk_global.h"
#include "jk_util.h"
#include "jk_map.h"
#include "jk_pool.h"
#include "jk_service.h"
#include "jk_worker.h"
#include "jk_ajp12_worker.h"
#include "jk_uri_worker_map.h"

#include "config.h"
#include "inifile.h"

#if !defined(DLLEXPORT)
#ifdef WIN32
#define DLLEXPORT __declspec(dllexport)
#else
#define DLLEXPORT
#endif
#endif

#define VERSION_JK_DSAPI_PLUGIN	"1.13.1"
#define VERSION_STRING		"Jakarta/Connector " JK_VERSTRING "/DSAPI/" VERSION_JK_DSAPI_PLUGIN
/* What we call ourselves */
#define FILTERDESC			"Apache Tomcat Interceptor (" VERSION_STRING ")"

#define SERVERDFLT			"Lotus Domino"
/* Registry location of configuration data */
#define REGISTRY_LOCATION	"Software\\Apache Software Foundation\\Jakarta Dsapi Redirector\\1.13.1"
/* Name of INI file relative to whatever the 'current' directory is when the filter is
 * loaded. Certainly on Linux this is the Domino data directory -- it seems likely that
 * it's the same on other platforms
 */
#define ININAME				"libtomcat.ini"

/* Names of registry keys/ini items that contain commands to start, stop Tomcat */
#define TOMCAT_START		"tomcat_start"
#define TOMCAT_STOP		"tomcat_stop"
#define TOMCAT_STARTSTOP_TO	30000				/* 30 seconds */

static int			initDone	= JK_FALSE;
static jk_uri_worker_map_t	*uw_map		= NULL;
static jk_logger_t		*logger		= NULL;

static int			logLevel	= JK_LOG_EMERG_LEVEL;
static jk_pool_t		cfgPool;

static const char 		*logFile;
static const char 		*workerFile;
static const char 		*workerMountFile;
static const char 		*tomcatStart;
static const char 		*tomcatStop;
static char			*crlf		= "\r\n";

static pthread_mutex_t 		initdone_mutex  = PTHREAD_MUTEX_INITIALIZER;

#if defined(JK_VERSION)
#if JK_VERSION >= JK_MAKEVERSION(1, 2, 7, 0)
#warning "This program has tested with JK 1.2.6 only!"
#elif JK_VERSION >= JK_MAKEVERSION(2, 0, 0, 0)
#error "This program is not JK2-compatible!"
#endif
#if JK_VERSION >= JK_MAKEVERSION(1, 2, 0, 1)
static jk_worker_env_t   worker_env;
#endif
#endif

#if defined( DEBUGLOG )
static char *methods[] = {
  "None",
  "HEAD",	
  "GET",	
  "POST",	
  "PUT",	
  "DELETE",
  "TRACE",
  "CONNECT",
  "OPTIONS",
  "UNKNOWN",
  "BAD"
};
#endif

/* Private context for request processing */
typedef struct private_ws
{
	jk_pool_t		p;

	/* These get passed in by Domino and are used to access various
	 * Domino methods and data.
	 */
	FilterContext		*context;
	FilterParsedRequest	*reqData;

	/* True iff the response headers have been sent
	 */
	int			responseStarted;

	/* Current pointer into and remaining size
	 * of request body data
	 */
	char			*reqBuffer;
	unsigned int		reqSize;

	pthread_mutex_t  ws_mutex;
	int		 finished;
	char	        *webName;
	char	        *lotusName;
	char	        *password;
	char	        *domSessId;
	char		*workerName;
        char            *uri;
	jk_ws_service_t *s;
} private_ws_t;

/*
 * These three functions are called back (indirectly) by
 * Tomcat during request processing. StartResponse() sends
 * the headers associated with the response.
 */
static int JK_METHOD StartResponse( jk_ws_service_t * s, int status, const char *reason,
				    const char *const *hdrNames,
				    const char *const *hdrValues, unsigned hdrCount );
/*
 * Read() is called by Tomcat to read from the request body (if any).
 */
static int JK_METHOD Read( jk_ws_service_t * s, void *b, unsigned l, unsigned *a );

/*
 * Write() is called by Tomcat to send data back to the client.
 */
static int JK_METHOD Write( jk_ws_service_t * s, const void *b, unsigned l );

static int ReadInitData( void );

#if defined( WIN32 ) && !defined( USE_INIFILE )
static const char *GetRegString( HKEY hkey, const char *key );
#endif

#ifdef DEBUGLOG
static void debuglog2( char *file, char *str, int len )
{
  va_list ap;

  FILE *f = NULL;

  if( f = fopen( file, "a" ) ) {
    fwrite(str, 1, len, f);
    fclose( f );
  }
}

static void debuglog( char *str, ... )
{
  va_list ap;

  FILE *f = NULL;

  if( f = fopen( DEBUGLOGFILE, "a" ) ) {
    va_start(ap, str);
    vfprintf(f, str, ap);
    va_end(ap);
    fclose( f );
  }
}
#else
#  define debuglog( ... )
#  define debuglog2( ... )
#endif

static unsigned int StartRequest( FilterContext *context );
static unsigned int ParsedRequest( FilterContext *context, FilterParsedRequest *reqData );
static unsigned int Authenticate( FilterContext* context, FilterAuthenticate *authData );
static unsigned int Authorized( FilterContext* context, FilterAuthorize *authData );
static unsigned int ProcessRequest( FilterContext *context, FilterMapURL *eventData );
static unsigned int EndRequest( FilterContext *context );

/* Notes SDK unix shared library entrypoint */
STATUS FAR PASCAL MainEntryPoint( void );

/*
 * Description:  Provide a main entry point for the Notes API to
 *               initialize inside of this shared library.  We need
 *               this entry point because we want to be able to lookup
 *               user information in the name and address book via the
 *               Notes SDK API.
 *
 * Input:  nothing
 * Output: nothing
 * Return: NOERROR
 */
STATUS FAR PASCAL MainEntryPoint (void)
{
   return NOERROR;
}

/* Case insentive memcmp() clone */
#ifdef HAVE_MEMICMP
#  define NoCaseMemCmp(ci, cj, l) _memicmp((void *) (ci), (void *) (cj), (l))
#else
static int NoCaseMemCmp(const char *ci, const char *cj, int len)
{
int cmp;

  for( ;len > 0;len-- ) {
    if( cmp = tolower(*ci++) - tolower(*cj++) )
      return cmp;
  }
  return 0;
}
#endif

/* Case insentive strcmp() clone
 */
#ifdef HAVE_STRICMP
#define NoCaseStrCmp(si, sj) _stricmp((void *) (si), (void *) (sj))
#else
static int NoCaseStrCmp(const char *si, const char *sj)
{
int cmp;

  do {
    if( cmp = tolower(*si) - tolower(*sj++) )
      return cmp;
  } while( *si++ );
  return cmp;
}
#endif

/*
 * Allocate space for a string given a start pointer and an end pointer
 * and return a pointer to the allocated, copied string.
 */
static char *MemDup(private_ws_t *ws, const char *start, const char *end)
{
  char *out = NULL;

  if( start != NULL && end != NULL && end > start ) {
    int len = end - start;
    out = jk_pool_alloc(&ws->p, len + 1);
    memcpy( out, start, len );
    out[len] = '\0';
  }
  return out;
}

/*
 * Case insensitive substring search.
 * str		string to search
 * slen		length of string to search
 * ptn		pattern to search for
 * plen		length of pattern
 * returns	1 if there's a match otherwise 0
 */
static int FindPathElem(const char *str, int slen, const char *ptn, int plen)
{
  const char *sp = str;

  while( slen >= plen ) {
	/* We're looking for a match for the specified string bounded by
	 * the start of the string, \ or / at the left and the end of the
	 * string, \ or / at the right. We look for \ as well as / on the
	 * suspicion that a Windows hosted server might accept URIs
	 * containing \.
	 */
    if( NoCaseMemCmp(sp, ptn, plen) == 0 &&
        (sp == str || *sp == '\\' || *sp == '/') &&
        (*sp == '\0' || *sp == '\\' || *sp == '/'))
      return 1;
    slen--;
    sp++;
  }
  return 0;
}

/*
 * Get the value of a server (CGI) variable as a string
 */
static int GetVariable(private_ws_t *ws, char *hdrName,
					 char *buf, DWORD bufsz, char **dest, const char *dflt)
{
	unsigned errID;

	if (ws->context->GetServerVariable(ws->context, hdrName, buf, bufsz, &errID))
		*dest = jk_pool_strdup(&ws->p, buf);
	else
		*dest = jk_pool_strdup(&ws->p, dflt);

	DEBUG(("%s = %s\n", hdrName, *dest));

	return JK_TRUE;
}

/* Get the value of a server (CGI) variable as an integer
 */
static int GetVariableInt( private_ws_t *ws, char *hdrName,
			   char *buf, DWORD bufsz, int *dest, int dflt )
{
	unsigned errID;

	if( ws->context->GetServerVariable(ws->context, hdrName, buf, bufsz, &errID) )
		*dest = atoi(buf);
	else
		*dest = dflt;

	DEBUG(("%s = %d\n", hdrName, *dest));

	return JK_TRUE;
}

/*
 * Get the value of a server (CGI) variable as an integer
 */
static int GetVariableBool(private_ws_t *ws, char *hdrName,
						char *buf, DWORD bufsz, int *dest, int dflt)
{
	unsigned errID;

	if (ws->context->GetServerVariable(ws->context, hdrName, buf, bufsz, &errID))
	{
		if (isdigit(buf[0]))
			*dest = atoi(buf) != 0;
		else if (NoCaseStrCmp(buf, "yes") == 0 || NoCaseStrCmp(buf, "on") == 0)
			*dest = 1;
		else
			*dest = 0;
	}
	else
	{
		*dest = dflt;
	}

	DEBUG(("%s = %d\n", hdrName, *dest));

	return JK_TRUE;
}

/*
 * A couple of utility macros to supply standard arguments to GetVariable() and
 * GetVariableInt().
 */
#define GETVARIABLE(name, dest, dflt)		GetVariable(ws, (name), workBuf, sizeof(workBuf), (dest), (dflt))
#define GETVARIABLEINT(name, dest, dflt)	GetVariableInt(ws, (name), workBuf, sizeof(workBuf), (dest), (dflt))
#define GETVARIABLEBOOL(name, dest, dflt)	GetVariableBool(ws, (name), workBuf, sizeof(workBuf), (dest), (dflt))

/*
 * Replacement for strcat() that updates a buffer pointer. It's
 * probably marginal, but this should be more efficient that strcat()
 * in cases where the string being concatenated to gets long because
 * strcat() has to count from start of the string each time.
 */
static void Append(char **buf, const char *str)
{
  int l = strlen( str );
  memcpy( *buf, str, l );
  (*buf)[ l ] = '\0';
  *buf += l;
}

/*
 * Start the response by sending any headers. Invoked by Tomcat.
 *
 * Andy: I don't particularly like the fact that this always allocates memory, but
 * perhaps jk_pool_alloc() is efficient.
 */
static int JK_METHOD StartResponse(jk_ws_service_t *s, int status, const char *reason,
					const char *const *hdrNames,
					const char *const *hdrValues, unsigned hdrCount)
{
int rc = JK_FALSE;

	debuglog( "TOMCAT [%08X] StartResponse invoked\n", ((private_ws_t *)s->ws_private)->context );
	jk_log(logger, JK_LOG_DEBUG, "Into jk_ws_service_t::StartResponse\n");
//	debuglog( "s->ws_private = %08X\n", s->ws_private );

	if (status < 100 || status > 1000)
	{
		jk_log(logger, JK_LOG_ERROR, "jk_ws_service_t::StartResponse, invalid status %d\n", status);
	}
	else if (s && s->ws_private) {
		private_ws_t *ws = s->ws_private;
                pthread_mutex_lock( &ws->ws_mutex );
		if( !ws->responseStarted ) {
			ws->responseStarted = JK_TRUE;
                        pthread_mutex_unlock( &ws->ws_mutex );
			char *hdrBuf;
			FilterResponseHeaders frh;
			unsigned errID;
			int ok;

			if( NULL == reason )
				reason = "";

			/* Build a single string containing all the headers
			 * because that's what Domino needs.
			 */
			if (hdrCount > 0) {
				unsigned i;
				unsigned hdrLen;
				char *bufp;

				for (i = 0, hdrLen = 3; i < hdrCount; i++)
					hdrLen += strlen(hdrNames[i]) + strlen(hdrValues[i]) + 4;

				hdrBuf = jk_pool_alloc(&ws->p, hdrLen);
				bufp = hdrBuf;

				for (i = 0; i < hdrCount; i++) {
					Append(&bufp, hdrNames[i]);
					Append(&bufp, ": ");
					Append(&bufp, hdrValues[i]);
					Append(&bufp, crlf);
				}

				Append(&bufp, crlf);
			}
			else {
				hdrBuf = crlf;
			}

			frh.responseCode = status;
			frh.reasonText = (char *) reason;
			frh.headerText = hdrBuf;

			DEBUG(("%d %s\n%s", status, reason, hdrBuf));

			/* Send the headers */
			debuglog( "kWriteResponseHeaders %d %s\n%s\n", status, reason, hdrBuf  );
			ok = ws->context->ServerSupport(ws->context, kWriteResponseHeaders, &frh, NULL, 0, &errID);
			if( !ok ) {
				jk_log(logger, JK_LOG_ERROR,
					   "jk_ws_service_t::StartResponse, ServerSupportFunction failed (rc=%d,err=%d)\n", ok, errID);
			}
			else
				rc = JK_TRUE;
		}
		else {
                  pthread_mutex_unlock( &ws->ws_mutex );
                  rc = JK_TRUE;
                }
	}
	else {
	  jk_log(logger, JK_LOG_ERROR, "jk_ws_service_t::StartResponse, NULL parameters\n");
	}

	debuglog( "TOMCAT [%08X] StartResponse finished (rc=%d)\n", ((private_ws_t *)s->ws_private)->context, rc );
	return rc;
}

/* */
static int JK_METHOD Read(jk_ws_service_t *s, void *bytes, unsigned len, unsigned *countp)
{
int rc = JK_FALSE;

	debuglog( "TOMCAT [%08X] Read(%p, %p, %u, %p)\n", ((private_ws_t *)s->ws_private)->context, s, bytes, len, countp );
	DEBUG(("Read(%p, %p, %u, %p)\n", s, bytes, len, countp));
	jk_log(logger, JK_LOG_DEBUG, "Into jk_ws_service_t::Read\n");

	if (s && s->ws_private && bytes && countp)
	{
		private_ws_t *ws = s->ws_private;

		/*
		 * Copy data from Domino's buffer. Although it seems slightly
		 * improbably we're believing that Domino always buffers the
		 * entire request in memory. Not properly tested yet.
		 */
                pthread_mutex_lock( &ws->ws_mutex );
		if (len > ws->reqSize) len = ws->reqSize;
		memcpy(bytes, ws->reqBuffer, len);
		ws->reqBuffer += len;
		ws->reqSize -= len;
                pthread_mutex_unlock( &ws->ws_mutex );
		*countp = len;
		rc = JK_TRUE;
	}
	else {
	  jk_log(logger, JK_LOG_ERROR, "jk_ws_service_t::Read, NULL parameters\n");
	}
	debuglog( "TOMCAT [%08X] Read finished (rc=%d)\n", ((private_ws_t *)s->ws_private)->context, rc );
	return rc;
}

/* */
static int JK_METHOD Write(jk_ws_service_t *s, const void *bytes, unsigned len)
{
int rc = JK_FALSE;

	debuglog( "TOMCAT [%08X] Write(%p, %p, %u)\n", ((private_ws_t *)s->ws_private)->context, bytes, len );
	DEBUG(("Write(%p, %p, %u)\n", s, bytes, len));
	jk_log(logger, JK_LOG_DEBUG, "Into jk_ws_service_t::Write\n");

//	debuglog( "s->ws_private = %08X\n", s->ws_private );
	if (s && s->ws_private && bytes) {
		private_ws_t *ws = s->ws_private;
		unsigned errID;
		int ok;
		DEBUG(("Writing %d bytes of content\n", len));

		/* Send the data */
		if (len > 0) {
			debuglog2( "/var/log/lotus/rawoutput.log", (char *)bytes, len );
			ok = ws->context->WriteClient( ws->context, (char *)bytes, len, 0, &errID );
			if( !ok ) {
				jk_log( logger, JK_LOG_ERROR,
					"jk_ws_service_t::StartResponse, ServerSupportFunction failed (rc=%d,err=%d)\n", ok, errID );
			}
			else
		  	  rc = JK_TRUE;
		}
		else
		  rc = JK_TRUE;
	}
	else {
	  jk_log(logger, JK_LOG_ERROR, "jk_ws_service_t::Write, NULL parameters\n");
	}

	debuglog( "TOMCAT [%08X] Write finished (rc=%d)\n", ((private_ws_t *)s->ws_private)->context, rc );
	return rc;
}

/* */
static int RunProg(const char *cmd)
{
#ifdef WIN32
    STARTUPINFO si;
    PROCESS_INFORMATION pi;

	ZeroMemory(&si, sizeof(si));
        si.cb			= sizeof(si);    // Start the child process.
	si.dwFlags		= STARTF_USESHOWWINDOW;
	si.wShowWindow	= SW_SHOWMAXIMIZED;

	if (!CreateProcess(NULL, (char *) cmd, NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi))
	{
		DWORD err = GetLastError();
		AddInLogMessageText("Command \"%s\" (error %u)", NOERROR, cmd, err);
		return FALSE;
	}

	if (WAIT_OBJECT_0 == WaitForSingleObject(pi.hProcess, TOMCAT_STARTSTOP_TO))
		return TRUE;

	AddInLogMessageText("Command \"%s\" didn't complete in time", NOERROR, cmd);
	return FALSE;

#else
	int err = system(cmd);
	if (0 == err) return 1;
	AddInLogMessageText("Command \"%s\" failed (error %d)", NOERROR, cmd, err);
	return 0;
#endif
}

/* Read parameters from the registry or ini-file */
static int ReadInitData(void)
{
	int ok = JK_TRUE;
	const char *v;

#ifdef USE_INIFILE
// Using an INIFILE

#define GETV(key) inifile_lookup(key)

	ERRTYPE e;

	if (e = inifile_read(&cfgPool, ININAME), ERRNONE != e)
	{
		AddInLogMessageText("Error reading: %s, %s", NOERROR, ININAME, ERRTXT(e));
		return JK_FALSE;
	}

#else
// Using the registry
#define GETV(key) GetRegString(hkey, key)

	HKEY hkey;
	long rc;

	rc = RegOpenKeyEx(HKEY_LOCAL_MACHINE, REGISTRY_LOCATION, (DWORD) 0, KEY_READ, &hkey);
	if (ERROR_SUCCESS != rc) return JK_FALSE;

#endif

#define GETVNB(tag, var) \
	var = GETV(tag); \
	if (NULL == var) \
	{ \
		AddInLogMessageText("%s not defined in %s", NOERROR, tag, ININAME); \
		ok = JK_FALSE; \
	}

	GETVNB(JK_LOG_FILE_TAG, logFile)
	GETVNB(JK_LOG_LEVEL_TAG, v);
	GETVNB(JK_WORKER_FILE_TAG, workerFile);
	GETVNB(JK_MOUNT_FILE_TAG, workerMountFile);

	logLevel = (NULL == v) ? 0 : jk_parse_log_level(v);

	tomcatStart	= GETV(TOMCAT_START);
	tomcatStop	= GETV(TOMCAT_STOP);

#if !defined( USE_INIFILE )
	RegCloseKey(hkey);
#endif

	return ok;
}

/*
 * Called when Domino loads the filter. Reads a load of config data from
 * the registry and elsewhere and displays a banner.
 */
DLLEXPORT unsigned int FilterInit(FilterInitData *filterInitData)
{
	debuglog( "DOMINO FilterInit Invoked\n" );
	jk_open_pool( &cfgPool, NULL, 0 ); /* empty pool for config data */

	if (!ReadInitData()) {
	  AddInLogMessageText("Error loading %s", NOERROR, FILTERDESC);
	  return kFilterError;
        }

	if (!jk_open_file_logger(&logger, logFile, logLevel))
		logger = NULL;

	if (NULL != tomcatStart && '\0' != *tomcatStart)
	{
		AddInLogMessageText("Attempting to start Tomcat: %s", NOERROR, tomcatStart);
		RunProg(tomcatStart);
	}

	jk_map_t *map = NULL;
	unsigned errID = JK_FALSE;
	debuglog( "Initializing worker map (%s) ", workerMountFile );
	if( map_alloc( &map ) ) {
		if( map_read_properties( map, workerMountFile ) )
			if (uri_worker_map_alloc(&uw_map, map, logger))
				errID = JK_TRUE;
			map_free(&map);
	}
	debuglog( "%s\n", errID ? "Ok" : "Error" );
	if( !errID ) {
	  AddInLogMessageText("Error loading %s", NOERROR, FILTERDESC);
	  return kFilterError;
	}
	filterInitData->eventFlags = kFilterStartRequest|kFilterParsedRequest|kFilterProcessRequest|
	                             /*kFilterAuthenticate|*/kFilterAuthorized|kFilterEndRequest;
	filterInitData->appFilterVersion = kInterfaceVersion;
	strcpy( filterInitData->filterDesc, FILTERDESC );

	// Banner
	AddInLogMessageText( "%s loaded", NOERROR, filterInitData->filterDesc );

	debuglog( "DOMINO FilterInit finished successfully\n" );
	return kFilterHandledEvent;
}

/*
 * Called when the filter is unloaded. Free various resources and
 * display a banner.
 */
DLLEXPORT unsigned int TerminateFilter(unsigned int reserved)
{
	if( initDone ) {
		uri_worker_map_free(&uw_map, logger);
		wc_close(logger);
		if (logger)
			jk_close_file_logger(&logger);
		initDone = JK_FALSE;
	}
	if(NULL != tomcatStop && '\0' != *tomcatStop) {
		AddInLogMessageText("Attempting to stop Tomcat: %s", NOERROR, tomcatStop);
		RunProg(tomcatStop);
	}
	AddInLogMessageText(FILTERDESC " unloaded", NOERROR);
	jk_close_pool(&cfgPool);
	return kFilterHandledEvent;
}

#if !defined( USE_INIFILE )
static const char *GetRegString(HKEY hkey, const char *key)
{
	DWORD type = 0;
	DWORD sz = 0;
	LONG rc;
	char *val;

	rc = RegQueryValueEx(hkey, key, (LPDWORD) 0, &type, NULL, &sz);
	if (rc != ERROR_SUCCESS || type != REG_SZ)
		return NULL;

	if (val = jk_pool_alloc(&cfgPool, sz), NULL == val)
		return NULL;

	rc = RegQueryValueEx(hkey, key, (LPDWORD) 0, &type, val, &sz);
	if (rc == ERROR_SUCCESS)
		return val;

	return NULL;
}
#endif

/*
 * The Event Notification function does the actual work of the filter.
 * Domino calls the Event Notification function whenever a particular event occurs during the processing of an http request.
 * When Domino calls the filter's Event Notification function it passes information about the request and the event being processed.
 * On each call the filter can decide to handle the event, with or without an error return, or decline to handle the event.
 *
 * The name of the event handling function must be HttpFilterProc for the DSAPI layer to find it when it loads the filter.
 *
 * pContext is a pointer to the FilterContext that contains information about the http request, server callback function pointers,
 * and a pointer to the filter's own data context. Refer to the Lotus C API Reference for further information on the FilterContext structure.
 *
 * eventType indicates which event is occurring.
 *
 * pEventData is a pointer to a structure associated with the event. There is a different event data structure for each event.
 *
 * All the exported functions return an unsigned int status code. The possible values of the status code are as follows:
 *
 *   kFilterHandledRequest signals that the http request has completely been processed. A response has already been sent to the client,
 *                         and the http stack must terminate the processing of the request.
 *   kFilterHandledEvent   signals that the event has been handled, but further processing is to take place to completely service
 *                         the http request.
 *   kFilterNotHandled     signals that the filter did not process the event.
 *   kFilterError          signals that the filter encountered an error while processing the event.
 *                         In this case the http stack is to stop processing the request, and send an appropriate error response to the user.
 *
 * Note that for the functions FilterInit, HttpFilterThreadInit, HttpFilterThreadTerm, and TerminateFilter,
 * the only possible return codes are kFilterHandledEvent and kFilterError. The function HttpFilterProc can return any of the status codes
 * when handling an event.
 */
DLLEXPORT unsigned int HttpFilterProc(FilterContext *context, unsigned int eventType, void *eventData)
{
unsigned rc = kFilterNotHandled;

        debuglog( "DOMINO [%08X] Filter event %04X invoked\n", context, eventType );
	switch (eventType) {
	case kFilterStartRequest:
		rc = StartRequest( context );
		break;
	case kFilterParsedRequest:
		rc = ParsedRequest(context, (FilterParsedRequest *) eventData);
		break;
	case kFilterAuthorized:
		rc = Authorized(context, (FilterAuthorize *) eventData);
		break;
	case kFilterProcessRequest:
                rc = ProcessRequest( context, (FilterMapURL *)eventData );
		break;
	case kFilterEndRequest:
		rc = EndRequest( context );
		break;
	case kFilterAuthenticate:
                Authenticate( context, (FilterAuthenticate *)eventData );
		rc = kFilterHandledEvent;
		break;
	case kFilterRawRequest:
	case kFilterAuthUser:
	case kFilterUserNameList:
	case kFilterMapURL:
	case kFilterResponse:
	case kFilterRawWrite:
	case kFilterPostTranslate:
	case kFilterRewriteURL:
		break;
	}
	return rc;
}

/*
 * Allocates copy of string argument in current Notes request context.
 * Notes API will automatically free allocated buffers after request has been processed
 */
static char *NewString( FilterContext *context, char *string )
{
  unsigned errID;
  char *s = context->AllocMem( context, strlen( string )+1, 0, &errID );
  if( s )
    strcpy( s, string );
  else {
    debuglog( "Can't allocate %d bytes of memory: err=%d\n", strlen( string )+1, &errID );
  }
  return s;
}

/*
 * This event is used to advise all filters currently loaded that an http request has been received,
 * and is about to be processed. A filter, in this step, can prepare itself for processing the request.
 * Typically in this step, the filters should allocate its own private context data, and required resources
 * necessary to handle the processing of a request.
 *
 * Any filter supporting this event should return the value, kFilterHandledEvent.
 */
static unsigned int StartRequest( FilterContext *context )
{
private_ws_t *ws;
jk_ws_service_t *s;
unsigned ok1, ok2, errID;
unsigned rc = kFilterNotHandled;
  
  debuglog( "DOMINO [%08X] StartRequest invoked\n", context );
  ws = context->AllocMem( context, sizeof( private_ws_t ), 0, &ok1 );
  pthread_mutex_init( &ws->ws_mutex, NULL );
  pthread_mutex_lock( &ws->ws_mutex );
//  debuglog( "AllocMem returns %8X (err=%d)\n", ws, ok1 );
  s = context->AllocMem( context, sizeof( jk_ws_service_t ), 0, &ok2 );
//  debuglog( "AllocMem returns %8X (err=%d)\n",  s, ok2 );
//  debuglog( "context=%08X;ws=%08X;s=%08X\n",context,ws,s );
  context->privateContext = ws;
  ws->context = context;
  ws->s = s;
  pthread_mutex_unlock( &ws->ws_mutex );

   /*
    * One time initialisation which is deferred so that we have the name of
    * the server software to plug into worker_env
    *
    * Semaphore required!
    */
  pthread_mutex_lock( &initdone_mutex );
  if( !initDone ) {
    int ok = JK_FALSE;
    debuglog( "Got the URI worker map\n" );
    jk_map_t *map = NULL;
    ok = JK_FALSE;
    debuglog( "About to allocate map " );
    if( map_alloc( &map ) ) {
      debuglog( "Ok\nAbout to read %s ", workerFile );
      if( map_read_properties( map, workerFile ) ) {
        char server[256];
        worker_env.uri_to_worker = uw_map;
        if( context->GetServerVariable( context, "SERVER_SOFTWARE", server, sizeof(server)-1, &errID ) ) {
          worker_env.server_name = jk_pool_strdup( &cfgPool, server );
          debuglog( "(Server name: %s) ", server );
        }
        else
          worker_env.server_name = SERVERDFLT;
        if( wc_open( map, &worker_env, logger ) )
          ok = JK_TRUE;
      }
      map_free( &map );
    }
    debuglog( "%s\n", ok ? "Ok" : "Error" );
    if( !ok ) {
      rc = kFilterError;
    }
    else
      initDone = JK_TRUE;
  }
  if( rc != kFilterError ) {
    jk_init_ws_service( s );
    s->ws_private = ws;
    rc = kFilterHandledEvent;
  }
  pthread_mutex_unlock( &initdone_mutex );

  debuglog( "DOMINO [%08X] StartRequest finished (rc=%d)\n", context, rc );
  return rc;
}

/*
 * The filter also define a Termination entry point.
 * Domino will call this function whenever the filter is about to be unloaded -
 * when the http stack process is shutting down. The filter use this function to clean up resources it allocated.
 * The filter must deallocate all global resources it allocated when its initialization routine was called.
 */
static unsigned int EndRequest( FilterContext *context )
{
  private_ws_t *ws = (private_ws_t *)context->privateContext;

  debuglog( "DOMINO [%08X] EndRequest invoked\n", context );
  pthread_mutex_lock( &ws->ws_mutex );
  jk_close_pool( &ws->p );
  pthread_mutex_unlock( &ws->ws_mutex );
  debuglog( "DOMINO [%08X] EndRequest finished\n", context );
  return kFilterHandledRequest;
}

/*
 * This function can be called after Authenticate event processed
 */
static void getAuthenticatedUserInfo( FilterContext *context, private_ws_t *ws )
{
int ok;
unsigned errID;
FilterAuthenticatedUser uinfo;

  pthread_mutex_lock( &ws->ws_mutex );
  memset( &uinfo, 0, sizeof uinfo );
  uinfo.fieldFlags = kCannonicalUserName | kWebUserName | kUserPassword;
  ok = context->ServerSupport( context, kGetAuthenticatedUserInfo, &uinfo, NULL, 0, &errID );
  if( ok == TRUE ) {
    debuglog( "[%08X] ServerSupport::kGetAuthenticatedUserInfo (%d/%d) -> %08X %08X %08X -> ", context, ok, errID,
              uinfo.pUserCannonicalName, uinfo.pWebUserName, uinfo.pUserPassword );

    if( uinfo.pUserCannonicalName ) {
      debuglog( "%s; ", uinfo.pUserCannonicalName );
      ws->lotusName = NewString( context, uinfo.pUserCannonicalName );
    }
    else
      ws->lotusName = NULL;
    if( uinfo.pWebUserName ) {
      debuglog( "%s; ", uinfo.pWebUserName );
      ws->webName = NewString( context, uinfo.pWebUserName );
    }
    else
      ws->webName = NULL;
    if( uinfo.pUserPassword ) {
      debuglog( "%s; ", uinfo.pUserPassword );
      ws->password = NewString( context, uinfo.pUserPassword );
    }
    else
      ws->password = NULL;
    debuglog( "\n" );
  }
  pthread_mutex_unlock( &ws->ws_mutex );
}

/*
 * Description:  This routine is called on a dsapi kFilterAuthenticate
 *               event.
 *
 * Input:  context     dsapi specification controls the format of data
 *         authData    password field contains the password to use for 
 *                     authentication userName field contains the name 
 *                     to authenticate foundInCache field is TRUE if 
 *                     user has been found in the cache and can be
 *                     authenticated on that basis.
 *
 * Return: kFilterNotHandled returns always for Notes-based authentication
 */
static unsigned int Authenticate(FilterContext* context, FilterAuthenticate *authData)
{
unsigned rc = kFilterNotHandled;
private_ws_t *ws = (private_ws_t *)context->privateContext;
  
  debuglog( "DOMINO [%08X] Authenticate invoked\n", context );
  if( !authData->foundInCache ) {
    if( NoCaseStrCmp( ws->workerName, "domino" ) ) { // our URI
      authData->authType = kAuthenticBasic;
      rc = kFilterHandledEvent;
    }
    else {
      authData->authType = kNotAuthentic;
      rc = kFilterNotHandled;
    }
  }
  debuglog( "DOMINO [%08X] Authenticate finished (rc=%d)\n", context, rc );
  return rc;
}

/*
 * Send a simple response. Used when we don't want to bother Tomcat,
 * which in practice means for various error conditions that we can
 * detect internally.
 */
static void SimpleResponse(FilterContext *context, int status, char *reason, char *body)
{
FilterResponseHeaders frh;
unsigned rc;
unsigned errID;
char hdrBuf[35];

	sprintf(hdrBuf, "Content-type: text/html%s%s", crlf, crlf);

	frh.responseCode = status;
	frh.reasonText = reason;
	frh.headerText = hdrBuf;

	rc = context->ServerSupport(context, kWriteResponseHeaders, &frh, NULL, 0, &errID);
	rc = context->WriteClient(context, body, strlen(body), 0, &errID);
}

/*
 * Return 1 iff the supplied string contains "web-inf" (in any case
 * variation. We don't allow URIs containing web-inf, although
 * FindPathElem() actually looks for the string bounded by path punctuation
 * or the ends of the string, so web-inf must appear as a single element
 * of the supplied URI
 */
static int isBadURI( const char *uri )
{
  static char *wi = "web-inf";
  return FindPathElem( uri, strlen( uri ), wi, strlen( wi ) );
}

/*
 * Called to reject a URI that contains the string "web-inf". We block
 * these because they may indicate an attempt to invoke arbitrary code.
 */
static unsigned int rejectBadURI(FilterContext *context)
{
  static char *msg = "<HTML><BODY><H1>Access is forbidden by Jakarta Tomcat Domino Connector</H1></BODY></HTML>";
  SimpleResponse(context, 403, "Forbidden", msg);
  return kFilterHandledRequest;
}

/*
 * This event occurs after the authentication phase has taken place and the user's group names list has been computed.
 * The filter can override the default implementation of the authorization phase and grant or deny access to the target resource.
 *
 * pEventData is a pointer to the structure, FilterAuthorize. It contains information about the target resource to serve.
 *
 * Note that the filter can get access to the authenticated user information by using the services via the ServerSupport callback
 * with the flag, kGetAuthenticatedUserInfo. This will allow the user to gain access to the user's authenticated name as well as
 * to his/her group names list.
 *
 * If the filter denies access to the target resource, it must send the appropriate response to the client, and set the field,
 * isAuthorized, in the FilterAuthorize structure to 0. It can then return either kFilterHandledRequest or kFilterHandledEvent.
 * The DSAPI layer will then signal the http stack to terminate the processing of the current request.
 */
static unsigned int Authorized( FilterContext *context, FilterAuthorize *authData )
{
unsigned rc = kFilterNotHandled;
private_ws_t *ws = (private_ws_t *)context->privateContext;

    debuglog( "DOMINO [%08X] Authorized invoked\n", context );
    if( NoCaseStrCmp( ws->workerName, "domino" ) ) { // our URI
      if( isBadURI( ws->uri ) ) {
        rejectBadURI( context );
        authData->isAuthorized = 0;
        rc = kFilterHandledRequest;
      }
      else {
        authData->isAuthorized = 1;
        rc = kFilterHandledEvent;
      }
    }
    debuglog( "DOMINO [%08X] Authorized finished (rc=%d)\n", context, rc );
    return rc;
}

/*
 * Given all the HTTP headers as a single string parse them into individual
 * name, value pairs. Called twice: once to work out how many headers there
 * are, then again to copy them.
 */
static int ParseHeaders(private_ws_t *ws, const char *hdrs, int hdrsz, jk_ws_service_t *s)
{
	int hdrCount = 0;
	const char *limit = hdrs + hdrsz;
	const char *name, *nameEnd;
	const char *value, *valueEnd;

	while (hdrs < limit)
	{
		/* Skip line *before* doing anything, cos we want to lose the first line which
		 * contains the request.
		 */
		while (hdrs < limit && (*hdrs != '\n' && *hdrs != '\r'))
			hdrs++;
		while (hdrs < limit && (*hdrs == '\n' || *hdrs == '\r'))
			hdrs++;

		if (hdrs >= limit)
			break;

		name = nameEnd = value = valueEnd = NULL;

		name = hdrs;
		while (hdrs < limit && *hdrs >= ' ' && *hdrs != ':')
			hdrs++;
		nameEnd = hdrs;

		if (hdrs < limit && *hdrs == ':')
		{
			hdrs++;
			while (hdrs < limit && (*hdrs == ' ' || *hdrs == '\t'))
				hdrs++;
			value = hdrs;
			while (hdrs < limit && *hdrs >= ' ')
				hdrs++;
			valueEnd = hdrs;
		}

		if (s->headers_names != NULL && s->headers_values != NULL)
		{
			s->headers_names[hdrCount]  = MemDup(ws, name, nameEnd);
			s->headers_values[hdrCount] = MemDup(ws, value, valueEnd);
			DEBUG(("%s = %s\n", s->headers_names[hdrCount], s->headers_values[hdrCount]));
#if 1
			if( !strcmp( s->headers_names[hdrCount], "Cookie" ) )
			  debuglog("%s = %s\n", s->headers_names[hdrCount], s->headers_values[hdrCount] );
#endif
		}
		hdrCount++;
	}

	return hdrCount;
}

static char *methodName[] = { "", "HEAD", "GET", "POST", "PUT", "DELETE", "TRACE", "CONNECT", "OPTIONS" };

/*
 * Set up all the necessary jk_* workspace based on the current HTTP request.
 */
static void InitService(private_ws_t *ws, jk_ws_service_t *s)
{
/* This is the only fixed size buffer left. It won't be overflowed
 * because the Domino API that reads into the buffer accepts a length
 * constraint, and it's unlikely ever to be exhausted because the
 * strings being will typically be short, but it's still aesthetically
 * troublesome.
 */
char workBuf[16 * 1024];
FilterRequest fr;
char *hdrs, *qp;
int hdrsz;
unsigned errID;
int hdrCount;
unsigned rc;

	rc = ws->context->GetRequest( ws->context, &fr, &errID );

	s->jvm_route		= NULL;
	s->start_response	= StartResponse;
	s->read			= Read;
	s->write		= Write;

	s->req_uri = jk_pool_strdup( &ws->p, fr.URL );
	s->query_string = NULL;
	if( qp = strchr(s->req_uri, '?') ) {
		*qp++ = '\0';
		if (strlen(qp))
			s->query_string = qp;
	}

	GETVARIABLE("SERVER_PROTOCOL", &s->protocol, "");
	GETVARIABLE("REMOTE_HOST", &s->remote_host, "");
	GETVARIABLE("REMOTE_ADDR", &s->remote_addr, "");
	GETVARIABLE("SERVER_NAME", &s->server_name, "");
	GETVARIABLEINT("SERVER_PORT", (int *)&s->server_port, 80);
	GETVARIABLE("SERVER_SOFTWARE", &s->server_software, SERVERDFLT);
	GETVARIABLEINT("CONTENT_LENGTH", (int *)&s->content_length, 0);

	/* SSL Support */
	GETVARIABLEBOOL("HTTPS", &s->is_ssl, 0);
//	debuglog( "HTTPS: %d\n", s->is_ssl );


	s->method = methodName[ws->reqData->requestMethod];
	s->headers_names	= NULL;
	s->headers_values	= NULL;
	s->num_headers		= 0;
	s->ssl_cert_len	= fr.clientCertLen;
	s->ssl_cert	= (char *)fr.clientCert;
	s->ssl_cipher	= NULL;		/* required by Servlet 2.3 Api */
	s->ssl_session	= NULL;

#if defined(JK_VERSION) && JK_VERSION >= MAKEVERSION(1, 2, 0, 1)
	s->ssl_key_size = -1;       /* required by Servlet 2.3 Api, added in jtc */
#endif

	if( s->is_ssl )	{
		int dummy;
		DEBUG(("SSL request\n"));

#if defined(JK_VERSION) && JK_VERSION >= MAKEVERSION(1, 2, 0, 1)
		/* Read the variable into a dummy variable: we do this for the side effect of
		 * reading it into workBuf.
		 */
		GETVARIABLEINT("HTTPS_KEYSIZE", &dummy, 0);
		debuglog( "HTTPS_KEYSIZE: %d\n", dummy );
		if (workBuf[0] == '[')
			s->ssl_key_size = atoi(workBuf+1);
#else
		(void) dummy;
#endif
	}

	/* Duplicate all the headers now */

	hdrsz = ws->reqData->GetAllHeaders(ws->context, &hdrs, &errID);
	DEBUG(("\nGot headers (length %d)\n--------\n%s\n--------\n\n", hdrsz, hdrs));

	s->headers_names =
	s->headers_values = NULL;
	hdrCount = ParseHeaders(ws, hdrs, hdrsz, s);
	DEBUG(("Found %d headers\n", hdrCount));
	s->num_headers = hdrCount;
	s->headers_names	= jk_pool_alloc(&ws->p, hdrCount * sizeof(char *));
	s->headers_values	= jk_pool_alloc(&ws->p, hdrCount * sizeof(char *));
	hdrCount = ParseHeaders(ws, hdrs, hdrsz, s);
}

static char *getDefaultWorker( char *uri )
{
regex_t preg;
int     ok;

  ok = regcomp( &preg, "^[/a-z0-9._-]+\\.nsf/?", REG_EXTENDED|REG_ICASE|REG_NOSUB );
  if( ok != 0 ) {
    char errbuf[ 4096 ];
    regerror( ok, &preg, errbuf, sizeof( errbuf ) );
    debuglog( "regcomp: %s\n", errbuf );
  }
  ok = regexec( &preg, uri, 0, NULL, 0 );
  printf( "regexec: %d\n", ok );
  regfree( &preg );
  if( ok == 0 )
    return "domino";
  return "ajp13";
}

/*
 * Handle an HTTP request. Works out whether Tomcat will be interested then either
 * despatches it to Tomcat or passes it back to Domino.
 */
static unsigned int ParsedRequest( FilterContext *context, FilterParsedRequest *reqData )
{
unsigned errID;
int ok;
FilterRequest fr;
unsigned rc = kFilterNotHandled;
private_ws_t *ws = (private_ws_t *)context->privateContext;
			    
  debuglog( "DOMINO [%08X] ParsedRequest invoked\n", context );
  ok = context->GetRequest( context, &fr, &errID );
  debuglog( "Method:        %s\n", methods[ fr.method ] );
  debuglog( "URL:           %s\n", fr.URL );
//	debuglog( "Version:       %s\n", fr.version );
  debuglog( "UserName:      %s\n", fr.userName ? fr.userName : "none" );
  debuglog( "Password:      %s\n", fr.password ? fr.password : "none" );
//	debuglog( "ClientCert:    %08X\n", fr.clientCert != NULL ? (void *)fr.clientCert : "none" );
//	debuglog( "ClientCertLen: %u\n", fr.clientCertLen );
//	debuglog( "ContentLen:    %u\n", fr.contentReadLen );

//	debuglog( "fr.URL=%s\n", fr.URL );

  if( reqData->requestMethod < 0 ||
      reqData->requestMethod >= sizeof(methodName) / sizeof(methodName[0])
    ) {
    debuglog( "Invalid method: %d\n", reqData->requestMethod );
    rc = kFilterError;
  }
  else if( fr.URL && strlen( fr.URL ) ) {
    char *uri = fr.URL;
    char *workerName, *qp, *turi;

    turi = strdup( uri );
    if( qp = strchr( turi, '?' ) ) *qp = '\0';
    workerName = (char *)map_uri_to_worker( uw_map, turi, logger );
    if( !workerName || !strncmp( workerName, "default=", 8 ) )
      workerName = getDefaultWorker( turi );
    free( turi );

    debuglog( "Worker for this URL is %s\n", workerName );

    if( NoCaseStrCmp( workerName, "ajp13" ) == 0 ) {
      ok = context->ServerSupport( context, kOwnsRequest, NULL, NULL, 0, &errID );
      debuglog( "[%08X] kOwnsRequest (%d/%d)\n", context, ok, errID );
      rc = kFilterHandledEvent;
    }

    if( workerName ) {
      void *buf = context->AllocMem( context, sizeof( jk_pool_atom_t ) * SMALL_POOL_SIZE, 0, &errID );
      jk_ws_service_t *s = ws->s;

      if( !NoCaseStrCmp( workerName, "domino" ) )
        rc = kFilterNotHandled;
      /* Go dispatch the call */
      if( rc != kFilterHandledRequest ) {
        pthread_mutex_lock( &ws->ws_mutex );
        jk_open_pool( &ws->p, buf, sizeof( jk_pool_atom_t ) * SMALL_POOL_SIZE );
        ws->responseStarted = JK_FALSE;
        ws->reqData         = reqData;
        ws->reqSize         = context->GetRequestContents( context, &ws->reqBuffer, &errID );
        ws->workerName      = NewString( context, workerName );
        ws->uri             = NewString( context, uri );
        s->pool             = &ws->p;
        InitService( ws, s );
        pthread_mutex_unlock( &ws->ws_mutex );
      }
    }
  }
  debuglog( "DOMINO [%08X] ParsedRequest finished (rc=%d)\n", context, rc );
  return rc;
}

/*
 * This is the last step in servicing an http request. This event can be used to override the default implementation
 * of the processing of the request. In this phase the response data is computed and sent to the client.
 * pEventData is a pointer to the structure, FilterMapURL.
 */
static unsigned int ProcessRequest( FilterContext *context, FilterMapURL *eventData )
{
unsigned rc = kFilterNotHandled;
private_ws_t *ws = (private_ws_t *)context->privateContext;
jk_ws_service_t *s = ws->s;
unsigned errID;

  debuglog( "DOMINO [%08X] ProcessRequest invoked\n", context );
//  if( ws->workerName && NoCaseStrCmp( ws->workerName, "domino" ) && !ws->finished ) {
    getAuthenticatedUserInfo( context, ws );
    if( ws->webName ) {
      s->remote_user = ws->webName; // memory allocated in server context
      debuglog( "remote_user: %s\n", s->remote_user );
    }
    jk_worker_t *worker = wc_get_worker_for_name( ws->workerName, logger );
    debuglog( "worker: %s (%08X)\n", ws->workerName, worker );
    jk_log( logger, JK_LOG_DEBUG, "HttpExtensionProc %s a worker for name %s\n",  worker ? "got" : "could not get", ws->workerName );
    if( worker ) {
      jk_endpoint_t *e = NULL;
      if( worker->get_endpoint( worker, &e, logger ) ) {
        int recover = JK_FALSE;
	debuglog( "e->service( %08X, %08X, %08X, %08X )\n", e, s, logger, &recover );
        if( e->service( e, s, logger, &recover ) ) {
          rc = kFilterHandledEvent;
          jk_log( logger, JK_LOG_DEBUG, "HttpExtensionProc service() returned OK\n" );
        }
        else {
          rc = kFilterError;
          jk_log( logger, JK_LOG_ERROR, "HttpExtensionProc service() failed\n" );
        }
//	debuglog( "doResponse: e->done: %08X, %08X\n", e, logger );
        e->done( &e, logger );
//	debuglog( "doResponse: done\n" );
	ws->finished = JK_TRUE;
	rc = kFilterHandledRequest;
      }
//      worker->destroy( &worker, logger );
    }
  debuglog( "DOMINO [%08X] ProcessRequest finished (rc=%d)\n", context, rc );
  return rc;
}
